import datetime
import pytest
import sys
import unittest
from django.core.management import call_command
from io import StringIO

from ..models import BlockedIP


@pytest.mark.django_db
class CommandsTest(unittest.TestCase):
    def setUp(self):
        BlockedIP.objects.all().delete()
        self.ip1 = "1.1.1.1"
        self.ip2 = "2.2.2.2"

    def tearDown(self):
        BlockedIP.objects.all().delete()

    def test_clean(self):
        two_days_ago = datetime.datetime.now() - datetime.timedelta(days=2)
        BlockedIP.objects.create(ip=self.ip1, cooldown=1, last_seen=two_days_ago)
        BlockedIP.objects.create(ip=self.ip2, cooldown=3, last_seen=two_days_ago)
        assert BlockedIP.objects.count() == 2
        # Running clean_blocklist with --dry-run doesn't delete one
        call_command("clean_blocklist", verbosity=0, dry_run=1)
        self.assertEqual(BlockedIP.objects.count(), 2)
        # Running clean_blocklist _does_ delete one
        call_command("clean_blocklist", verbosity=0)
        self.assertEqual(BlockedIP.objects.count(), 1)
        # The one that should still be there, is still there
        self.assertTrue(BlockedIP.objects.filter(ip=self.ip2).exists())

    def test_add(self):
        call_command("update_blocklist", self.ip1, verbosity=0)
        self.assertTrue(BlockedIP.objects.filter(ip=self.ip1).exists())

    def test_remove(self):
        ip = "4.4.4.4"
        BlockedIP.objects.create(ip=ip)
        call_command("remove_from_blocklist", ip, verbosity=0)
        self.assertEqual(BlockedIP.objects.filter(ip=ip).count(), 0)

    def test_add_invalid(self):
        sys.stdout = (out := StringIO())
        bad_ip = "foo"
        call_command("update_blocklist", bad_ip, verbosity=0)
        self.assertIn("Invalid", out.getvalue())

    def test_update_changes_stored_values(self):
        BlockedIP.objects.create(ip=self.ip1, reason="R1", cooldown=1)
        call_command("update_blocklist", self.ip1, reason="R2", cooldown=2)
        entry = BlockedIP.objects.get(ip=self.ip1)
        self.assertEqual(entry.reason, "R2")
        self.assertEqual(entry.cooldown, 2)

    def test_update_creation_message(self):
        """Newly created entry should be reported"""
        sys.stdout = (out := StringIO())
        call_command("update_blocklist", self.ip1)
        self.assertIn("Created", out.getvalue())

    def test_update_output_names_changed_fields(self):
        """If we update reason or cooldown, output should say so"""
        sys.stdout = (out := StringIO())
        BlockedIP.objects.create(ip=self.ip1, reason="R1", cooldown=1)
        call_command("update_blocklist", self.ip1, cooldown=2)
        self.assertIn("Updated cooldown for", out.getvalue())
        call_command("update_blocklist", self.ip1, reason="R2")
        self.assertIn("Updated reason for", out.getvalue())
        call_command("update_blocklist", self.ip1, reason="R3", cooldown=3)
        self.assertIn("Updated reason and cooldown", out.getvalue())

    def test_update_converts_cooldown_to_int(self):
        """Logic inside the update command needs cooldown to be cast to int"""
        sys.stdout = (out := StringIO())
        BlockedIP.objects.create(ip=self.ip1, cooldown=1)
        call_command("update_blocklist", self.ip1, cooldown="1")
        self.assertNotIn("Updated cooldown", out.getvalue())

    def test_update_no_message_if_no_change(self):
        """If we 'update' with the existing values, IP shouldn't show in output"""
        sys.stdout = (out := StringIO())
        BlockedIP.objects.create(ip=self.ip1, reason="R1", cooldown=1)
        call_command("update_blocklist", self.ip1, reason="R1", cooldown=1)
        self.assertNotIn(self.ip1, out.getvalue())

    def test_report(self):
        sys.stdout = (out := StringIO())
        today = datetime.datetime.now()
        yesterday = today - datetime.timedelta(days=1)
        BlockedIP.objects.create(ip=self.ip1, first_seen=yesterday, last_seen=today)
        BlockedIP.objects.create(ip=self.ip2, first_seen=yesterday, last_seen=today, tally=240)
        call_command("report_blocklist")
        result = out.getvalue()
        self.assertIn(f"{self.ip1} -- 0 blocks", result)
        self.assertIn("10 per hour", result)

    def test_reason_report(self):
        sys.stdout = (out := StringIO())
        reasons = ["A", "B"]
        for n, reason in enumerate(reasons):
            BlockedIP.objects.create(ip=f"{n}.{n}.{n}.{n}", reason=reason, tally=1)
        call_command("report_blocklist", reason="A")
        result = out.getvalue()
        # import pdb;pdb.set_trace()
        assert BlockedIP.objects.count() == 2
        self.assertIn("Entries in blocklist: 1", result)
        self.assertIn("        1 |         1 | A", result)



"""Update blocklist with new IPs, or set metadata on existing IPs."""
import logging

from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand
from django.core.validators import validate_ipv46_address

from ...models import BlockedIP
from ...utils import COOLDOWN


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("ips", nargs="+", type=str, help="IPs (space-separated)")
        parser.add_argument(
            "--cooldown",
            help=f"Days with no requests before IP is dropped from blocklist (default: {COOLDOWN})",
        )
        parser.add_argument("--reason", help="'reason' field value for these IPs", default="")

    help = __doc__

    def handle(self, *args, **options):
        ips = options.get("ips")
        cooldown = options.get("cooldown")
        reason = options.get("reason")
        for ip in ips:
            try:
                validate_ipv46_address(ip)
            except ValidationError:
                print(f"Invalid IP: {ip}")
                continue
            entry, created = BlockedIP.objects.get_or_create(ip=ip)
            updated = []
            if reason and entry.reason != reason:
                entry.reason = reason
                updated.append("reason")
            if cooldown and entry.cooldown != (cooldown := int(cooldown)):
                entry.cooldown = cooldown
                updated.append("cooldown")
            if updated or created:
                entry.save()
                summary = "Created entry" if created else f"Updated {' and '.join(updated)}"
                print(f"{summary} for {ip}")



from django.apps import AppConfig


class Config(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_blocklist"
    # These can be overridden via settings.BLOCKLIST_CONFIG
    defaults = {
        "cache-ttl": 60,
        "cooldown": 7,
        "denial-template": "Your IP address {ip} has been blocked. Try again in {cooldown} days.",
    }



# Generated by Django 3.2.13 on 2022-07-02 15:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_blocklist", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="blockedip",
            name="tally",
            field=models.IntegerField(
                default=0, help_text="Number of times this IP has been blocked since first_seen"
            ),
        ),
    ]



import datetime

from django.contrib import admin
from django.http import HttpResponseRedirect

from .models import BlockedIP


@admin.display(description="Reason")
def reason_truncated(entry: BlockedIP) -> str:
    return entry.reason[:20] + ("..." if len(entry.reason) > 20 else "")


@admin.display(description="Cooldown")
def cooldown(entry: BlockedIP) -> str:
    return f"{entry.cooldown} days"


@admin.display(description="Days left")
def days_left(entry: BlockedIP) -> str:
    if entry.last_seen:
        remaining = f"{entry.cooldown - (datetime.datetime.today() - entry.last_seen).days}"
    else:
        remaining = ""
    return remaining


@admin.action(permissions=["view"])
def look_up_first_selected_IP(modeladmin, request, queryset):
    obj = queryset[0]
    return HttpResponseRedirect(f"https://whatismyipaddress.com/ip/{obj.ip}")


class BlockedIPAdmin(admin.ModelAdmin):
    list_display = ["ip", "first_seen", "last_seen", "tally", cooldown, days_left, reason_truncated]
    list_filter = ["first_seen", "last_seen", "cooldown", "reason"]
    search_fields = ["ip", "reason"]
    actions = [look_up_first_selected_IP]

    class Meta:
        model = BlockedIP


admin.site.register(BlockedIP, BlockedIPAdmin)



"""Print summary information about the data in the blocklist."""
import datetime
from collections import Counter
from operator import itemgetter
from typing import Iterable, Tuple

from django.contrib.humanize.templatetags.humanize import intcomma
from django.core.management.base import BaseCommand, CommandError
from django.db.models import Sum

from ...models import BlockedIP


class Command(BaseCommand):
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument(
            "--reason",
            default=[],
            action="append",
            help="Restrict report to IPs with this reason (muliple reasons can be passed)",
        )

    help = __doc__

    def handle(self, *args, **options):
        entries = BlockedIP.objects.all()
        if selected_reasons := options.get("reason"):
            entries = entries.filter(reason__in=selected_reasons)
        if entries.count() == 0:
            raise CommandError("No BlockedIP objects for report.")
        _grand_tally = entries.aggregate(Sum("tally"))["tally__sum"]
        print(f"Total blocks of listed IPs: {intcomma(_grand_tally)}")
        print(f"Entries in blocklist: {intcomma(entries.count())}")
        _one_day_ago = datetime.datetime.now() - datetime.timedelta(days=1)
        print(f"Active in last 24 hours: {intcomma(entries.filter(last_seen__gte=_one_day_ago).count())}")
        print(
            f"Stale (added over 24h ago, not seen since): {intcomma(entries.filter(tally=1, last_seen__lt=_one_day_ago).count())}"
        )
        print()
        print_roster("Most recent", entries.exclude(tally=0).order_by("-last_seen")[:5])
        print_roster("Most active", entries.filter(tally__gt=0).order_by("-tally"), activity_calc=True)
        longest_lived = None
        how_long = datetime.timedelta(0)
        for entry in entries:
            active_period = entry.last_seen - entry.first_seen
            if active_period > how_long:
                longest_lived, how_long = entry, active_period
        if longest_lived is not None:
            print(f"Longest lived:\n{longest_lived.verbose_str()}")
        if stats := reason_stats():
            print("Tally and IP count by reason\n----------------------------")
            for tally, ip_count, reason in stats:
                if not selected_reasons or reason in selected_reasons:
                    print(f"{tally:9} | {ip_count:9} | {reason}")


def print_roster(title: str, queryset, activity_calc: bool = False) -> None:
    print(f"{title}:")
    activity = {}
    for perp in queryset:
        # For each IP, either print a line or calculate the rate, depending on activity_calc
        if activity_calc:
            days = max(1, (perp.last_seen - perp.first_seen).days)
            per_hour = perp.tally / days / 24
            activity[perp] = per_hour
        else:
            print(perp.verbose_str())
    if activity_calc:
        most_active = sorted(activity.items(), key=itemgetter(1), reverse=True)[:5]
        for perp, per_hour in most_active:
            print(f"{perp.verbose_str()} -- {round(per_hour)} per hour")
    print()


def reason_stats() -> Iterable[Tuple[int, int, str]]:
    """Gather ip-count and grand tally stats for each reason"""
    stats = Counter(BlockedIP.objects.exclude(reason="").values_list("reason"))
    tuples = []
    for reason_datum in stats.items():
        reason = str(reason_datum[0][0])
        ip_count = reason_datum[1]
        tally = BlockedIP.objects.filter(reason=reason).aggregate(tally=Sum("tally"))["tally"]
        tuples.append((tally, ip_count, reason))
    return sorted(tuples, key=itemgetter(0), reverse=True)
